// This file is taken from [`chili`]

use std::{
    cell::UnsafeCell,
    ptr::NonNull,
    sync::{
        Arc,
        atomic::{AtomicU8, AtomicU32, Ordering},
    },
    thread,
};

enum State {
    Pending,
    Waiting,
    Ready,
    Taken,
}

pub use werkzeug::sync::Parker;

use crate::queue::Queue;

#[derive(Debug)]
#[repr(C)]
struct Channel<T = ()> {
    state: AtomicU8,
    /// Can only be written only by the `Receiver` and read by the `Sender` if
    /// `state` is `State::Waiting`.
    waiting_thread: NonNull<Parker>,
    /// Can only be written only by the `Sender` and read by the `Receiver` if
    /// `state` is `State::Ready`.
    val: UnsafeCell<Option<Box<thread::Result<T>>>>,
}

impl<T> Channel<T> {
    fn new(waiting_thread: NonNull<Parker>) -> Self {
        Self {
            state: AtomicU8::new(State::Pending as u8),
            waiting_thread,
            val: UnsafeCell::new(None),
        }
    }
}

#[derive(Debug)]
pub struct Receiver<T = ()>(Arc<Channel<T>>);

impl<T: Send> Receiver<T> {
    pub fn is_empty(&self) -> bool {
        self.0.state.load(Ordering::Acquire) != State::Ready as u8
    }

    pub fn sender(&self) -> Sender<T> {
        Sender(self.0.clone())
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn wait(&self) {
        loop {
            match self.0.state.compare_exchange(
                State::Pending as u8,
                State::Waiting as u8,
                Ordering::AcqRel,
                Ordering::Acquire,
            ) {
                Ok(_) => {
                    // SAFETY:
                    // The `waiting_thread` is set to the current thread's parker
                    // before we park it.
                    unsafe {
                        let thread = self.0.waiting_thread.as_ref();
                        thread.park();
                    }

                    // we might have been woken up because of a shared job.
                    // In that case, we need to check the state again.
                    if self
                        .0
                        .state
                        .compare_exchange(
                            State::Waiting as u8,
                            State::Pending as u8,
                            Ordering::AcqRel,
                            Ordering::Acquire,
                        )
                        .is_ok()
                    {
                        continue;
                    } else {
                        // The state was changed to `State::Ready` by the `Sender`, so we can return.
                        break;
                    }
                }
                Err(state) if state == State::Ready as u8 => {
                    // The channel is ready, so we can return immediately.
                    return;
                }
                _ => {
                    panic!("Receiver is already waiting or consumed.");
                }
            }
        }
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn poll(&self) -> Option<thread::Result<T>> {
        if self
            .0
            .state
            .compare_exchange(
                State::Ready as u8,
                State::Taken as u8,
                Ordering::AcqRel,
                Ordering::Acquire,
            )
            .is_ok()
        {
            unsafe { Some(self.take()) }
        } else {
            None
        }
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn recv(self) -> thread::Result<T> {
        self.wait();

        // SAFETY:
        // To arrive here, either `state` is `State::Ready` or the above
        // `compare_exchange` succeeded, the thread was parked and then
        // unparked by the `Sender` *after* the `state` was set to
        // `State::Ready`.
        //
        // In either case, this thread now has unique access to `val`.
        assert_eq!(
            self.0.state.swap(State::Taken as u8, Ordering::Acquire),
            State::Ready as u8
        );

        unsafe { self.take() }
    }

    unsafe fn take(&self) -> thread::Result<T> {
        let result = unsafe { (*self.0.val.get()).take().map(|b| *b).unwrap() };

        result
    }
}

#[derive(Debug)]
#[repr(transparent)]
pub struct Sender<T = ()>(Arc<Channel<T>>);

impl<T: Send> Sender<T> {
    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn send(self, val: thread::Result<T>) {
        // SAFETY:
        // Only this thread can write to `val` and none can read it
        // yet.
        unsafe {
            *self.0.val.get() = Some(Box::new(val));
        }

        if self.0.state.swap(State::Ready as u8, Ordering::AcqRel) == State::Waiting as u8 {
            // SAFETY:
            // A `Receiver` already wrote its thread to `waiting_thread`
            // *before* setting the `state` to `State::Waiting`.
            unsafe {
                let thread = self.0.waiting_thread.as_ref();
                thread.unpark();
            }
        }
    }

    pub unsafe fn parker(&self) -> &Parker {
        unsafe { self.0.waiting_thread.as_ref() }
    }

    /// The caller must ensure that this function or `send` are only ever called once.
    pub unsafe fn send_as_ref(&self, val: thread::Result<T>) {
        // SAFETY:
        // Only this thread can write to `val` and none can read it
        // yet.
        unsafe {
            *self.0.val.get() = Some(Box::new(val));
        }

        if self.0.state.swap(State::Ready as u8, Ordering::AcqRel) == State::Waiting as u8 {
            // SAFETY:
            // A `Receiver` already wrote its thread to `waiting_thread`
            // *before* setting the `state` to `State::Waiting`.
            unsafe {
                let thread = self.0.waiting_thread.as_ref();
                thread.unpark();
            }
        }
    }
}

pub fn channel<T: Send>(thread: NonNull<Parker>) -> (Sender<T>, Receiver<T>) {
    let channel = Arc::new(Channel::new(thread));

    (Sender(channel.clone()), Receiver(channel))
}
